import math

# Instead of having fiat core as a dependency for this one function that doesnt use any other functions from fiat core,
# the function was copied here (26/9/2024, 661e0f2b2d6396346140316412c5957bc10eb03b) from https://github.com/Deltares/Delft-FIAT/blob/master/src/fiat/models/calc.py 

def calc_rp_coef(
    rp: list | tuple,
):
    """
    Calculates coefficients used to compute the EAD as a linear function of
    the known damages.

    Parameters
    ----------
    rp : list or tuple of int
        Return periods T1 … Tn for which damages are known.

    Returns
    -------
    list of float
        Coefficients a1, …, an used to compute the EAD as a linear function of the known damages.

    Notes
    -----
    In which D(f) is the damage, D, as a function of the frequency of exceedance, f.
    In order to compute this EAD, function D(f) needs to be known for the entire range of frequencies.
    Instead, D(f) is only given for the n frequencies as mentioned in the table above. 
    So, in order to compute the integral above, some assumptions need to be made for function D(f):
    (i)   For f > f1 the damage is assumed to be equal to 0.
    (ii)  For f < fn, the damage is assumed to be equal to Dn.
    (iii) For all other frequencies, the damage is estimated from log-linear interpolation 
          between the known damages and frequencies.
    """
    # Step 1: Compute frequencies associated with T-values.
    _rp = sorted(rp)
    idxs = [_rp.index(n) for n in rp]
    rp_u = sorted(rp)
    rp_l = len(rp_u)

    f = [1 / n for n in rp_u]
    lf = [math.log(1 / n) for n in rp_u]

    if rp_l == 1:
        return f

    # Step 2:
    c = [(1 / (lf[idx] - lf[idx + 1])) for idx in range(rp_l - 1)]

    # Step 3:
    G = [(f[idx] * lf[idx] - f[idx]) for idx in range(rp_l)]

    # Step 4:
    a = [
        (
            (1 + c[idx] * lf[idx + 1]) * (f[idx] - f[idx + 1])
            + c[idx] * (G[idx + 1] - G[idx])
        )
        for idx in range(rp_l - 1)
    ]
    b = [
        (c[idx] * (G[idx] - G[idx + 1] + lf[idx + 1] * (f[idx + 1] - f[idx])))
        for idx in range(rp_l - 1)
    ]

    # Step 5:
    alpha = [
        b[0]
        if idx == 0
        else f[idx] + a[idx - 1]
        if idx == rp_l - 1
        else a[idx - 1] + b[idx]
        for idx in range(rp_l)
    ]

    return [alpha[idx] for idx in idxs]




from abc import ABC, abstractmethod
from collections import Counter
from enum import Enum
from pathlib import Path
from typing import Optional, Union

import geopandas as gpd
import pandas as pd
import numpy as np
import pandas as pd
import shapely.geometry as geom
import math
from fiat_toolbox import FiatColumns, get_fiat_columns

def generate_polygon(point, shape_type, diameter):
    """
    Generate a polygon of a specified shape and diameter centered at a given point.

    Parameters
    ----------
    point (shapely.geometry.Point): The center point of the polygon.
    shape_type (str): The type of shape to generate. Must be one of 'circle', 'square', or 'triangle'.
    diameter (float): The diameter of the shape.

    Returns
    -------
    shapely.geometry.Polygon: The generated polygon.

    Raises
    ------
    ValueError: If the shape_type is not one of 'circle', 'square', or 'triangle'.
    """
    if shape_type == "circle":
        return point.buffer(diameter / 2)
    elif shape_type == "square":
        half_side = diameter / 2
        return geom.Polygon(
            [
                (point.x - half_side, point.y - half_side),
                (point.x + half_side, point.y - half_side),
                (point.x + half_side, point.y + half_side),
                (point.x - half_side, point.y + half_side),
            ]
        )
    elif shape_type == "triangle":
        height = (math.sqrt(3) / 2) * diameter
        return geom.Polygon(
            [
                (point.x, point.y - height / 2),
                (point.x - diameter / 2, point.y + height / 2),
                (point.x + diameter / 2, point.y + height / 2),
            ]
        )
    else:
        raise ValueError(
            "Invalid shape type. Choose from 'circle', 'square', or 'triangle'."
        )
    
def check_extension(out_path, ext):
    """
    Checks if the file extension of the given path matches the specified extension.

    Parameters:
    out_path (str or Path): The path to the file.
    ext (str): The expected file extension (including the dot, e.g., '.txt').

    Raises:
    ValueError: If the file extension of out_path does not match the specified ext.
    """
    out_path = Path(out_path)
    if out_path.suffix != ext:
        raise ValueError(
            f"File extention given: '{out_path.suffix}' does not much the file format specified: {ext}."
        )

def mode(my_list):
    """
    Calculate the mode(s) of a list.

    The mode is the value that appears most frequently in a data set. If there are multiple values with the same highest frequency, all of them are returned in a sorted list.

    Parameters:
    my_list (list): A list of elements to find the mode of.

    Returns:
    list: A sorted list of the mode(s) of the input list.
    """
    ct = Counter(my_list)
    max_value = max(ct.values())
    return sorted(key for key, value in ct.items() if value == max_value)


class Footprints:

    def __init__(self, 
                 footprints: gpd.GeoDataFrame,
                 field_name: Optional[str] = "BF_FID",
                 fiat_columns: Optional[FiatColumns] = None,
                 fiat_version: Optional[str] = "0.2",
                 ):
        """
        Initialize the Footprints object.
        Parameters:
        -----------
        footprints : gpd.GeoDataFrame
            A GeoDataFrame containing the footprint geometries.
        field_name : Optional[str], default "BF_FID"
            The name of the field to be used as the index. Must be present in the footprints columns.
        fiat_columns : Optional[FiatColumns], default None
            An object containing the column naming format. If None, the default format for the specified fiat_version will be used.
        fiat_version : Optional[str], default "0.2"
            The version of the FIAT format to use for column naming.
        Raises:
        -------
        AttributeError
            If the specified field_name is not present in the footprints columns.
        ValueError
            If the values in the specified field_name are not unique.
        """
        # Check if field name is present
        if field_name not in footprints.columns:
            raise AttributeError(f"field_name= '{field_name}' is not in footprints columns.")
        # Check if indices are unique
        if not footprints[field_name].is_unique:
            raise ValueError(f"Values in the field '{field_name}' are not unique.")
        # Save attributes
        footprints = footprints.set_index(field_name)
        self.footprints = footprints
        self.field_name = field_name
        # Get column naming format
        if fiat_columns is None:
            self.fiat_columns = get_fiat_columns(fiat_version=fiat_version)
        else:
            self.fiat_columns = fiat_columns
        
    def aggregate(self, 
                  objects: Union[gpd.GeoDataFrame, pd.DataFrame], 
                  field_name: Optional[str] = None, 
                  drop_no_footprints: Optional[bool] = False, 
                  no_footprints_shape: str = "triangle", 
                  no_footprints_diameter: float = 10.
                  ):
        """
        Aggregate objects based on a specified field name that connects to the footprints unique id. If objects has 
        spatial information, it can be used for objects without footprint connections to make standard shape footprints.
        Parameters:
        -----------
        objects : Union[gpd.GeoDataFrame, pd.DataFrame]
            The spatial objects to be aggregated. Can be a GeoDataFrame or DataFrame.
        field_name : Optional[str], default=None
            The field name to merge and aggregate objects on. If not provided, it defaults to the instance's field_name.
        drop_no_footprints : Optional[bool], default=False
            If True, objects without footprints will be dropped. If False, they will be assigned a default shape.
        no_footprints_shape : str, default="triangle"
            The shape to assign to objects without footprints. Options include "triangle", "circle", etc.
        no_footprints_diameter : float, default=10.0
            The diameter of the shape to assign to objects without footprints.
        Raises:
        -------
        AttributeError
            If the specified field_name is not found in the columns of the provided objects.
        Returns:
        --------
        None
            The aggregated results are stored in the instance's `aggregated_results` attribute.
        """        
        # Merge based on "field_name" column
        if field_name is None: # if field_name is not provided assume it is the same as the footprints one
            field_name = self.field_name
        if field_name not in objects.columns:
            raise AttributeError(f"'{field_name}' not found columns of the provided objects.")
        gdf = self.footprints.merge(objects.drop(columns="geometry", errors="ignore"), on=field_name, how="outer")

        # Remove the building footprints without any object attached
        gdf = gdf.loc[~gdf[self.fiat_columns.object_id].isna()]
        gdf[self.fiat_columns.object_id] = gdf[self.fiat_columns.object_id].astype(int) # ensure that object ids are interpreted correctly as integers

        # Get column names per type
        columns = self._get_column_names(gdf)
        
        for col in columns["string"]:
            gdf[col] = gdf[col].astype(str)
        
        # Aggregate objects with the same "field_name"
        count = np.unique(gdf[field_name], return_counts=True)
        multiple_bffid = count[0][count[1] > 1][:-1]

        # First, combine the Primary Object Type and Object ID
        bffid_object_mapping = {}
        bffid_objectid_mapping = {}
        for bffid in multiple_bffid:
            all_objects = gdf.loc[gdf[field_name] == bffid, self.fiat_columns.primary_object_type].to_numpy()
            all_object_ids = gdf.loc[gdf[field_name] == bffid, self.fiat_columns.object_id].to_numpy()
            bffid_object_mapping.update(
                {bffid: "_".join(mode(all_objects))}
            )
            bffid_objectid_mapping.update(
                {bffid: "_".join([str(x) for x in all_object_ids])}
            )
        gdf.loc[gdf[field_name].isin(multiple_bffid), self.fiat_columns.primary_object_type] = gdf[field_name].map(
            bffid_object_mapping
        )
        gdf.loc[gdf[field_name].isin(multiple_bffid), self.fiat_columns.object_id] = gdf[field_name].map(
            bffid_objectid_mapping
        )

        # Aggregated results using different functions based on type of output
        mapping = {}
        for name in columns["string"]:
            mapping[name] = pd.Series.mode
        for name in columns["depth"]:
            mapping[name] = "mean"
        for name in columns["damage"]:
            mapping[name] = "sum"

        agg_cols = columns["string"] + columns["depth"] + columns["damage"]

        df_groupby = (
            gdf.loc[gdf[field_name].isin(multiple_bffid), [field_name] + agg_cols]
            .groupby(field_name)
            .agg(mapping)
        )

        # Replace values in footprints file
        for agg_col in agg_cols:
            bffid_aggcol_mapping = dict(zip(df_groupby.index, df_groupby[agg_col]))
            gdf.loc[gdf[field_name].isin(multiple_bffid), agg_col] = gdf[field_name].map(
                bffid_aggcol_mapping
            )

        # Drop duplicates
        gdf = gdf.drop_duplicates(subset=[field_name])
        gdf = gdf.reset_index(drop=True)
        exposure = [self.fiat_columns.object_id, "geometry"] + agg_cols
        gdf = gdf[exposure]

        for col in columns["string"]:
            for ind, val in enumerate(gdf[col]):
                if isinstance(val, np.ndarray):
                    gdf.loc[ind, col] = str(val[0])
        
        # Add extra footprints
        extra_footprints = []
        
        # If point object don't have a footprint reference assume a shape
        if not drop_no_footprints and "geometry" in objects.columns:
            no_footprint_objects = self._no_footprint_points_to_polygons(objects, no_footprints_shape, no_footprints_diameter)
            no_footprint_objects = no_footprint_objects[[self.fiat_columns.object_id, "geometry"] + agg_cols].to_crs(gdf.crs)
            extra_footprints.append(no_footprint_objects)
        
        # Add objects which are already described by a polygon
        if "geometry" in objects.columns:
            footprint_objects = self._find_footprint_objects(objects)[[self.fiat_columns.object_id, "geometry"] + agg_cols].to_crs(gdf.crs)
            extra_footprints.append(footprint_objects)
        
        # Combine
        gdf = pd.concat([gdf] + extra_footprints, axis=0)

        self.aggregated_results = gdf
    
    def calc_normalized_damages(self):
        """
        Calculate normalized damages for the aggregated results.
        This method calculates the normalized damages per type and total damage percentage
        for the given aggregated results based on the run type. The results are stored back
        in the `aggregated_results` attribute.
        For "event" run type:
        - Calculates the percentage damage per type and total damage percentage.
        For "risk" run type:
        - Calculates the total damage percentage and risk (Expected Annual Damage) percentage.
        The calculated percentages are rounded to 2 decimal places and stored in new columns
        in the GeoDataFrame.
        Attributes:
            aggregated_results (GeoDataFrame): The aggregated results containing damage data.
            run_type (str): The type of run, either "event" or "risk".
        Returns:
            None
        """
        gdf = self.aggregated_results
        # Calculate normalized damages per type
        value_cols = gdf.columns[gdf.columns.str.startswith(self.fiat_columns.max_potential_damage)].tolist()
        
        # Only for event type calculate % damage per type
        if self.run_type == "event":
            dmg_cols = gdf.columns[gdf.columns.str.startswith(self.fiat_columns.damage)].tolist()
            # Do per type
            for dmg_col in dmg_cols:
                new_name = dmg_col + " %"
                name = dmg_col.split(self.fiat_columns.damage)[1]
                gdf[new_name] = gdf[dmg_col] / gdf[self.fiat_columns.max_potential_damage + name] * 100
                gdf[new_name] = gdf[new_name].round(2)
            
            # Do total
            gdf["Total Damage %"] = gdf[self.fiat_columns.total_damage] / gdf.loc[:, value_cols].sum(axis=1) * 100
            gdf["Total Damage %"] = gdf["Total Damage %"].round(2).fillna(0)
            
        elif self.run_type == "risk":
            tot_dmg_cols = gdf.columns[gdf.columns.str.startswith(self.fiat_columns.total_damage)].tolist()
            for tot_dmg_col in tot_dmg_cols:
                new_name = tot_dmg_col + " %"
                gdf[new_name] = gdf[tot_dmg_col] / gdf.loc[:, value_cols].sum(axis=1) * 100
                gdf[new_name] = gdf[new_name].round(2)
            gdf["Risk (EAD) %"] = gdf[self.fiat_columns.risk_ead] / gdf.loc[:, value_cols].sum(axis=1) * 100
            gdf["Risk (EAD) %"] = gdf["Risk (EAD) %"].round(2).fillna(0)
        
        self.aggregated_results = gdf
    
    def write(self, output_path: Union[str, Path]):
        """
        Writes the aggregated results to a file.

        Parameters:
        output_path (Union[str, Path]): The path where the output file will be saved. 
                                        It can be a string or a Path object.

        Returns:
        None
        """
        self.aggregated_results.to_file(output_path, driver="GPKG")
    
    def _get_column_names(self, gdf):
        """
        Extracts and categorizes column names from a GeoDataFrame based on predefined criteria.
        Parameters:
        gdf (GeoDataFrame): The input GeoDataFrame containing the columns to be categorized.
        Returns:
        dict: A dictionary with keys 'string', 'depth', and 'damage', each containing a list of column names.
            - 'string': Columns that are strings and will be aggregated.
            - 'depth': Columns related to inundation depth (only if total damage is present).
            - 'damage': Columns related to damage, including potential damage and total damage.
        Raises:
        ValueError: If neither 'total_damage' nor 'ead_damage' columns are present in the GeoDataFrame.
        """
        
        # Get string columns that will be aggregated
        string_columns = [self.fiat_columns.primary_object_type] +  [col for col in gdf.columns if self.fiat_columns.aggregation_label in col]

        # Get type of run and columns
        if self.fiat_columns.total_damage in gdf.columns:
            self.run_type = "event"
            # If event save inundation depth
            depth_columns = [col for col in gdf.columns if self.fiat_columns.inundation_depth in col]
            # And all type of damages
            damage_columns = [
                col
                for col in gdf.columns
                if self.fiat_columns.damage in col and self.fiat_columns.max_potential_damage not in col and self.fiat_columns.damage_function not in col
            ]
            damage_columns.append(self.fiat_columns.total_damage)
        elif self.fiat_columns.risk_ead in gdf.columns:
            self.run_type = "risk"
            depth_columns = []
            # For risk only save total damage per return period and EAD
            damage_columns = [col for col in gdf.columns if self.fiat_columns.total_damage in col]
            damage_columns.append(self.fiat_columns.risk_ead)
        else:
            raise ValueError(
                f"The is no {self.fiat_columns.total_damage} or {self.fiat_columns.risk_ead} column in the results."
            )
        # add the max potential damages
        pot_damage_columns = [col for col in gdf.columns if self.fiat_columns.max_potential_damage in col]
        damage_columns = pot_damage_columns + damage_columns
        
        # create mapping dictionary
        dict = {"string": string_columns,
                "depth": depth_columns,
                "damage": damage_columns,
                }
        
        return dict
    
    def _find_footprint_objects(self, objects):
        """
        Identifies and returns objects that have a footprint.

        This method filters the input objects to find those that do not have a 
        value in the specified field (self.field_name) and have a geometry type 
        of "Polygon".

        Parameters:
        objects (GeoDataFrame): A GeoDataFrame containing spatial objects with 
                                geometries and attributes.

        Returns:
        GeoDataFrame: A GeoDataFrame containing objects that have a footprint 
                      (i.e., objects with missing values in the specified field 
                      and a geometry type of "Polygon").
        """
        buildings_with_footprint = objects[
            (objects[self.field_name].isna()) & (objects.geometry.type.isin(["Polygon", "MultiPolygon"]))
        ]
        return buildings_with_footprint
    
    def _no_footprint_points_to_polygons(self, objects, shape, diameter):
        """
        Converts point geometries of buildings without footprints to polygon geometries.
        This method identifies buildings that do not have footprint information and converts their point geometries 
        to polygon geometries based on the specified shape and diameter.
        Args:
            objects (GeoDataFrame): A GeoDataFrame containing building geometries and attributes.
            shape (str): The shape of the polygon to generate (e.g., 'circle', 'square').
            diameter (float): The diameter of the polygon to generate.
        Returns:
            GeoDataFrame or None: A GeoDataFrame with updated polygon geometries for buildings without footprints,
                                  or None if there are no such buildings.
        """
        # Find buildings with no footprint connected
        buildings_without_footprint = objects[
            (objects[self.field_name].isna()) & (objects.geometry.type == "Point")
        ]
        if len(buildings_without_footprint) > 1:
            init_crs = buildings_without_footprint.crs
            buildings_without_footprint = buildings_without_footprint.to_crs(
                buildings_without_footprint.estimate_utm_crs()
            )
            shape_type = shape
            diameter = diameter

            # Transform points to shapes
            buildings_without_footprint["geometry"] = buildings_without_footprint[
                "geometry"
            ].apply(lambda point: generate_polygon(point, shape_type, diameter))
            buildings_without_footprint = buildings_without_footprint.to_crs(init_crs)
            
        return buildings_without_footprint
    


import logging
from pathlib import Path
from typing import Dict, List, Union

import tomli
import validators
from PIL import Image
from plotly.graph_objects import Bar, Figure, Pie
from plotly.subplots import make_subplots

from fiat_toolbox.infographics.infographics_interface import IInfographicsParser
from fiat_toolbox.metrics_writer.fiat_read_metrics_file import MetricsFileReader


class InfographicsParser(IInfographicsParser):
    """Class for creating the infographic"""

    def __init__(
        self,
        scenario_name: str,
        metrics_full_path: Union[Path, str],
        config_base_path: Union[Path, str],
        output_base_path: Union[Path, str],
        logger: logging.Logger = logging.getLogger(__name__),
    ) -> None:
        """Initialize the InfographicsParser

        Parameters
        ----------
        scenario_name : str
            The name of the scenario
        metrics_full_path : Union[Path, str]
            The path to the metrics file
        config_base_path : Union[Path, str]
            The path to the config folder
        output_base_path : Union[Path, str]
            The path to the output folder
        """

        # Save the scenario name
        self.scenario_name = scenario_name

        # Convert the metrics path to a Path object
        if isinstance(metrics_full_path, str):
            metrics_full_path = Path(metrics_full_path)
        self.metrics_full_path = metrics_full_path

        # Convert the config path to a Path object
        if isinstance(config_base_path, str):
            config_base_path = Path(config_base_path)
        self.config_base_path = config_base_path

        # Convert the output path to a Path object
        if isinstance(output_base_path, str):
            output_base_path = Path(output_base_path)
        self.output_base_path = output_base_path

        self.logger = logger

    def _get_impact_metrics(self) -> Dict:
        """Get the impact metrics for a scenario

        Returns
        -------
        Dict
            The impact metrics for the scenario
        """

        # Check if the metrics file exists
        if not Path.exists(self.metrics_full_path):
            raise FileNotFoundError(
                f"Metrics file not found at {self.metrics_full_path}"
            )

        # Read configured metrics
        metrics = (
            MetricsFileReader(self.metrics_full_path)
            .read_metrics_from_file()
            .to_dict()["Value"]
        )

        # Return the metrics
        return metrics

    @staticmethod
    def _get_pies_dictionary(
        pie_chart_config_path: Union[str, Path], metrics: Dict
    ) -> Dict:
        """Get a dictionary which contains the configuration and data for the pie charts

        Parameters
        ----------
        pie_chart_config_path : Union[str, Path]
            The path to the pie chart configuration file
        metrics : Dict
            The impact metrics for the scenario

        Returns
        -------
        Dict
            The dictionary which contains the configuration and data for the pie charts
        """

        # Convert the pie chart configuration path to a Path object
        if isinstance(pie_chart_config_path, str):
            pie_chart_config_path = Path(pie_chart_config_path)

        # Check if the pie chart configuration file exists
        if not Path.exists(pie_chart_config_path):
            raise FileNotFoundError(
                f"Infographic configuration file not found at {pie_chart_config_path}"
            )

        # Initialize the pie chart dictionary
        pie_dict = {}
        with open(pie_chart_config_path, mode="rb") as fc:
            # Read the pie chart configuration
            pie_chart_config = tomli.load(fc)

            # Check if the charts are defined
            if "Charts" not in pie_chart_config:
                raise KeyError("Charts not found in pie chart configuration file")

            # Read the charts configuration
            for key, value in pie_chart_config["Charts"].items():
                pie_dict[value["Name"]] = {}
                pie_dict[value["Name"]]["Name"] = value["Name"]
                if "Image" in value:
                    pie_dict[value["Name"]]["Image"] = value["Image"]
                pie_dict[value["Name"]]["Values"] = []
                pie_dict[value["Name"]]["Colors"] = []
                pie_dict[value["Name"]]["Labels"] = []

            # Check if the categories are defined
            if "Categories" not in pie_chart_config:
                raise KeyError("Categories not found in pie chart configuration file")

            # Read the categories configuration
            categorie_dict = {}
            for key, value in pie_chart_config["Categories"].items():
                categorie_dict[value["Name"]] = {}
                categorie_dict[value["Name"]]["Name"] = value["Name"]
                categorie_dict[value["Name"]]["Color"] = value["Color"]
                if "Image" in value:
                    categorie_dict[value["Name"]]["Image"] = value["Image"]

            # Check if the slices are defined
            if "Slices" not in pie_chart_config:
                raise KeyError("Slices not found in pie chart configuration file")

            # Read the configuration for the seperate pie slices
            for key, value in pie_chart_config["Slices"].items():
                pie_dict[value["Chart"]]["Values"].append(
                    float(metrics[value["Query"]])
                )
                pie_dict[value["Chart"]]["Labels"].append(value["Category"])
                pie_dict[value["Chart"]]["Colors"].append(
                    categorie_dict[value["Category"]]["Color"]
                )
                if "Image" in categorie_dict[value["Category"]]:
                    if "Image" not in pie_dict[value["Chart"]]:
                        pie_dict[value["Chart"]]["Image"] = []
                    pie_dict[value["Chart"]]["Image"].append(
                        categorie_dict[value["Category"]]["Image"]
                    )

            # Check if the "Other" category is defined
            if "Other" in pie_chart_config:
                pie_dict["Other"] = {}
                for key, value in pie_chart_config["Other"].items():
                    pie_dict["Other"][key] = value

        return pie_dict

    @staticmethod
    def _figures_list_to_html(
        figs,
        file_path: Union[str, Path] = "infographics.html",
    ):
        """Save a list of plotly figures in an HTML file

        Parameters
        ----------
            figs : list[plotly.graph_objects.Figure]
                List of plotly figures to be saved. As it is currently implemented,
                the first figure will be the top half of the HTML file, the second
                figure will be the bottom left and the third figure will be the bottom right.
                If the list is shorter than 3, the remaining figures will be empty.
            file_path : Union[str, Path], optional
                Path to the HTML file, by default "infographics.html"

        Returns
        -------
            None

        Raises
        ------
            ValueError
                If the number of figures too large
            FileExistsError
                If the file_path already exists
            ValueError
                If the file_path is not a .html file

        """

        # Check if the number of figures is correct
        if len(figs) > 3:
            raise ValueError("Only 3 figures are allowed")

        # Convert the file_path to a Path object
        if isinstance(file_path, str):
            file_path = Path(file_path)

        # Check if the file_path already exists
        if Path.exists(file_path):
            raise FileExistsError(f"File already exists at {file_path}")

        # Check if the file_path is correct
        if file_path.suffix != ".html":
            raise ValueError(f"File path must be a .html file, not {file_path}")

        # Create the directory if it does not exist
        if not Path.exists(file_path.parent):
            file_path.parent.mkdir(parents=True)

        # Write the html to the file
        with open(file_path, "w", encoding="utf-8") as infographics:
            figure1_html = (
                figs[0].to_html(config={'displayModeBar': False}).split("<body>")[1].split("</body>")[0]
                if len(figs) > 0
                else ""
            )
            figure2_html = (
                figs[1].to_html(config={'displayModeBar': False}).split("<body>")[1].split("</body>")[0]
                if len(figs) > 1
                else ""
            )
            figure3_html = (
                figs[2].to_html(config={'displayModeBar': False}).split("<body>")[1].split("</body>")[0]
                if len(figs) > 2
                else ""
            )

            infographics.write(
                f"""
                <!DOCTYPE html>
                <html>
                    <head>
                        <style>
                        .container {{
                            display: flex;
                            flex-direction: column;
                            align-items: center;
                            justify-content: center;  # Center the plots vertically
                        }}
                        .top-half, .bottom {{
                            display: flex;
                            justify-content: center;
                            align-items: center;  # Center the plots vertically within their divs
                            width: 100%;
                        }}
                        .top-half {{
                            width: 100%;
                        }}
                        .bottom {{
                            flex-direction: row;
                        }}
                        .bottom-left, .bottom-right {{
                            width: 50%;
                            align-items: center;  # Center the plots vertically within their divs
                        }}
                    </style>
                    </head>
                    <body>
                        <div class="container">
                            <div class="top-half">
                                {figure1_html}
                            </div>
                            <div class="bottom">
                                <div class="bottom-left">
                                    {figure2_html}
                                </div>
                                <div class="bottom-right">
                                    {figure3_html}
                                </div>
                            </div>
                        </div>
                    </body>
                </html>
                """
            )

    @staticmethod
    def _check_image_source(
        img: str, image_folder_path: str = None, return_image: bool = True
    ) -> Union[str, Image.Image, None]:
        """Check if the image source is a url or a local path. If so, return the image source. If not, return None

        Parameters
        ----------
            img : str
                The image source
            image_folder_path : str, optional
                The path to the image folder, by default None
            return_image : bool, optional
                Whether to return the image or the path to the image, by default True returns the image

        Returns
        -------
            Union[str, Image, None]
                The image source or None if the image source is not a url or a local path
        """
        # Check if the image is a url. If so, add the image to the pie chart
        if validators.url(img):
            # Add the pie chart image
            return img
        elif image_folder_path and "{image_path}" in img:
            path = Path(img.replace("{image_path}", str(image_folder_path)))
            if Path.exists(path):
                if return_image:
                    return Image.open(path)
                else:
                    return str(path)
            else:
                return None
        else:
            path = Path(img)
            # Check if the given path is an absolute path
            if Path.exists(path):
                if return_image:
                    return Image.open(path)
                else:
                    return str(path)
            else:
                return None

    @staticmethod
    def _add_info_button(
        fig: Figure, plot_info: str, img: str, img_path: str, scale: float
    ) -> Figure:
        """Add an info button to a plotly figure

        Parameters
        ----------
            fig : plotly.graph_objects.Figure
                The plotly figure to which the info button should be added
            plot_info : str
                The text that should be shown when the info button is clicked
            img : str
                The image source
            img_path : str
                The path to the image folder
            scale : float
                The scale of the image

        Returns
        -------
            plotly.graph_objects.Figure
                The plotly figure with the info button
        """

        # Check if the image source is a url or a local path
        img_source = InfographicsParser._check_image_source(img, img_path)

        # Add an image
        fig.add_layout_image(
            {
                "source": img_source,
                "xref": "paper",
                "yref": "paper",
                "x": 1,
                "y": 1,
                "sizex": scale,
                "sizey": scale,
                "xanchor": "center",
                "yanchor": "top",
            }
        )

        # Add a hover label to the image
        fig.add_annotation(
            xref="paper",
            yref="paper",
            x=1,
            y=1,
            text="   ",
            xanchor="center",
            yanchor="top",
            showarrow=False,
            font={"size": 16, "color": "LightSeaGreen"},
            hovertext=plot_info,
        )

        return fig

    @staticmethod
    def _get_pie_chart_figure(data: Dict, **kwargs):
        """Create a pie chart figure from the pie chart dictionary, usually created by _get_pies_dictionary

        Parameters
        ----------
            data : Dict
                The pie chart dictionary
            **title : str, optional
                The title of the pie chart, by default ""
            **title_font_size : int, optional
                The font size of the title, by default 25
            **subtitle_font_size : int, optional
                The font size of the subtitle, by default 20
            **image_scale : float, optional
                The scale of the image, by default 0.2
            **numbers_font : int, optional
                The font size of the numbers, by default 20
            **legend_font_size : int, optional
                The font size of the legend, by default 20
            **legend_orientation : str, optional
                The orientation of the legend, by default "h"
            **yanchor : str, optional
                The y anchor of the legend, by default "bottom"
            **y : float, optional
                The y position of the legend, by default 1
            **xanchor : str, optional
                The x anchor of the legend, by default "center"
            **x : float, optional
                The x position of the legend, by default 0.5
            **image_path : Union[str, Path], optional
                The path to the image folder, by default None
            **plot_width : int, optional
                The width of the plot in pixels, by default len(data)*200
            **plot_height : int, optional
                The height of the plot in pixels, by default 500
            **plot_info : str, optional
                The plot info, by default ""
            **plot_info_img : str, optional
                The plot info image, by default ""
            **plot_info_scale : float, optional
                The scale of the plot info image, by default 0.2

        Returns
        -------
            Figure
                The pie chart figure
        """

        # Remove the "Other" category if it exists
        if "Other" in data:
            data.pop("Other")

        # Get the title and legend configuration with default values
        title = kwargs.get("title", "")
        title_font_size = kwargs.get("title_font_size", 25)
        subtitle_font_size = kwargs.get("subtitle_font_size", 20)
        image_scale = kwargs.get("image_scale", 0.2)
        numbers_font = kwargs.get("numbers_font", 20)
        legend_font_size = kwargs.get("legend_font_size", 20)
        legend_orientation = kwargs.get("legend_orientation", "h")
        yanchor = kwargs.get("yanchor", "bottom")
        y = kwargs.get("y", 1)
        xanchor = kwargs.get("xanchor", "center")
        x = kwargs.get("x", 0.5)
        image_path = kwargs.get("image_path", None)
        plot_width = kwargs.get("plot_width", len(data) * 200)
        plot_height = kwargs.get("plot_height", 500)
        plot_info = kwargs.get("plot_info", "")
        plot_info_img = kwargs.get("plot_info_img", "")
        plot_info_scale = kwargs.get("plot_info_scale", 0.2)

        # Create the pie chart figure
        fig = make_subplots(
            rows=1,
            cols=len(data),
            specs=[[{"type": "domain"}] * len(data)],
            horizontal_spacing=0.2 / len(data),
            vertical_spacing=0,
        )

        # Add the pie chart to the figure
        for idx, (key, value) in enumerate(data.items()):
            # Create single pie chart
            trace = Pie(
                values=value["Values"],
                labels=value["Labels"],
                hole=0.6,
                hoverinfo="label+percent+name+value",
                textinfo="none",
                name=value["Name"],
                direction="clockwise",
                sort=False,
                marker={
                    "line": {"color": "#000000", "width": 2},
                    "colors": value["Colors"],
                },
            )

            # Add the pie chart to the figure
            fig.add_trace(trace, row=1, col=idx + 1)

            # Get the center of the pie chart (domain)
            domain_center_x = sum(fig.get_subplot(row=1, col=idx + 1).x) / 2
            domain_center_y = sum(fig.get_subplot(row=1, col=idx + 1).y) / 2

            # Add the title annotation
            fig.add_annotation(
                x=domain_center_x,
                y=1,
                text=f"{ value['Name'] } <br> ",
                font={
                    "size": subtitle_font_size,
                    "family": "Verdana",
                    "color": "black",
                },
                xanchor="center",
                yanchor="middle",
                showarrow=False,
            )

            # Add the image to the pie chart
            img_source = InfographicsParser._check_image_source(
                value["Image"], image_path
            )

            if img_source:
                fig.add_layout_image(
                    {
                        "source": img_source,
                        "sizex": image_scale,
                        "sizey": image_scale,
                        "x": domain_center_x,
                        "y": domain_center_y + 0.05,
                        "xanchor": "center",
                        "yanchor": "middle",
                        "visible": True,
                    }
                )

            # Add the sum of all slices to the pie chart
            fig.add_annotation(
                x=domain_center_x,
                y=domain_center_y - 0.05,
                text="{:,.0f}".format(sum(value["Values"])),
                font={"size": numbers_font, "family": "Verdana", "color": "black"},
                xanchor="center",
                yanchor="top",
                showarrow=False,
            )

        # Final update for the layout
        fig.update_layout(
            title_text=title,
            title_font={"size": title_font_size, "family": "Verdana", "color": "black"},
            title_x=0.5,
            width=plot_width,  # Set the width in pixels
            height=plot_height,  # Set the height in pixels
            legend={
                "orientation": legend_orientation,
                "yanchor": yanchor,
                "y": y,
                "xanchor": xanchor,
                "x": x,
                "itemclick": False,
                "itemdoubleclick": False,
                "font": {
                    "size": legend_font_size,
                    "family": "Verdana",
                    "color": "black",
                },
            },
        )

        # Add an info button
        fig = InfographicsParser._add_info_button(
            fig, plot_info, plot_info_img, image_path, plot_info_scale
        )

        # Update the layout images
        fig.update_layout_images()

        return fig

    @staticmethod
    def _get_bar_chart_figure(data: Dict, **kwargs):
        """Create a bar chart figure from the bar chart dictionary, usually created by _get_pies_dictionary

        Parameters
        ----------
        data : Dict
            The bar chart dictionary
        **title : str, optional
            The title of the bar chart, by default ""
        **yaxis_title : str, optional
            The title of the y axis of the bar chart, by default ""
        **title_font_size : int, optional
            The font size of the title, by default 25
        **subtitle_font_size : int, optional
            The font size of the subtitle, by default 20
        **image_scale : float, optional
            The scale of the image, by default 0.2
        **numbers_font : int, optional
            The font size of the numbers, by default 20
        **image_path : Union[str, Path], optional
            The path to the image folder, by default None
        **plot_width : int, optional
            The width of the plot in pixels, by default len(data)*200
        **plot_height : int, optional
            The height of the plot in pixels, by default 500
        **plot_info : str, optional
            The plot info, by default ""
        **plot_info_img : str, optional
            The plot info image, by default ""
        **plot_info_scale : float, optional
            The scale of the plot info image, by default 0.2

        Returns
        -------
        Figure
            The bar chart figure
        """

        # Get the title and legend configuration with default values
        title = kwargs.get("title", "")
        yaxis_title = kwargs.get("yaxis_title", "")
        title_font_size = kwargs.get("title_font_size", 25)
        subtitle_font_size = kwargs.get("subtitle_font_size", 20)
        image_scale = kwargs.get("image_scale", 0.2)
        numbers_font = kwargs.get("numbers_font", 20)
        image_path = kwargs.get("image_path", None)
        plot_width = kwargs.get("plot_width", 600)
        plot_height = kwargs.get("plot_height", 500)
        plot_info = kwargs.get("plot_info", "")
        plot_info_img = kwargs.get("plot_info_img", "")
        plot_info_scale = kwargs.get("plot_info_scale", 0.2)

        # Remove the "Other" category if it exists
        if "Other" in data:
            data.pop("Other")

        # Create the pie chart figure
        fig = make_subplots(
            rows=1,
            cols=len(data),
            specs=[[{"type": "domain"}] * len(data)],
            horizontal_spacing=0.2 / len(data),
            vertical_spacing=0,
        )

        # Add the bar chart to the figure
        for idx, (key, chart) in enumerate(data.items()):
            # Create single bar chart
            for i, (value, color, label, image) in enumerate(
                zip(chart["Values"], chart["Colors"], chart["Labels"], chart["Image"])
            ):
                # Add bar to the figure
                fig.add_trace(
                    Bar(
                        x=[label],
                        y=[int(value)],
                        marker={
                            "color": color,
                            "line": {"color": "black", "width": 2},
                        },
                        hoverinfo="x+y",
                    )
                )

                # Add annotation with the value on top of the bar
                fig.add_annotation(
                    x=label,
                    y=value + 0.1,
                    text="{:,.0f}".format(value),
                    showarrow=False,
                    font={
                        "color": "black",  # Set text color to white
                        "size": numbers_font,  # Set text size to 14
                    },
                    xref="x",
                    yref="y",
                    yanchor="bottom",
                )

                # Add the image to the bar chart
                img_source = InfographicsParser._check_image_source(image, image_path)

                if img_source:
                    # Calculate paper coordinates for the image
                    x_paper = (i + 0.5) / len(chart["Labels"])

                    # Add image below the bar
                    fig.add_layout_image(
                        {
                            "source": img_source,
                            "xref": "paper",
                            "yref": "paper",
                            "x": x_paper,
                            "y": -0.1 * image_scale,
                            "sizex": image_scale,
                            "sizey": image_scale,
                            "xanchor": "center",
                            "yanchor": "top",
                            "sizing": "contain",
                            "visible": True,
                        }
                    )

            # Add a new annotation to serve as the x-axis title
            fig.add_annotation(
                {
                    "x": 0.5,
                    "y": -image_scale,  # Adjust this value to move the title up or down
                    "xref": "paper",
                    "yref": "paper",
                    "showarrow": False,
                    "text": "Road users",  # X-axis title
                    "font": {"size": numbers_font},  # Adjust font size as needed
                    "xanchor": "center",
                    "yanchor": "top",
                }
            )

            # Update layout for better visualization
            fig.update_layout(
                title={
                    "text": title,
                    "x": 0.5,  # Center the main title
                },
                title_font={
                    "size": title_font_size,
                    "family": "Verdana",
                    "color": "black",
                },  # Adjust font size as needed
                xaxis={
                    "showticklabels": False,  # Hide x-axis labels
                },
                yaxis={
                    "title": yaxis_title,
                    "showgrid": False,  # Hide y-axis grid lines
                    "zeroline": False,  # Hide zero line
                    "showticklabels": False,  # Hide y-axis labels
                    "title_font": {
                        "size": subtitle_font_size,
                        "family": "Verdana",
                        "color": "black",
                    },  # Adjust font size as needed
                },
                width=plot_width,  # Set the width in pixels
                height=plot_height,  # Set the height in pixels
                showlegend=False,  # Remove legend
                plot_bgcolor="rgba(0, 0, 0, 0)",  # Set plot background color to transparent
                bargap=0.2,  # Set the gap between bars (adjust as needed)
                bargroupgap=0.5,  # Set the gap between bar groups (adjust as needed)
            )

        # Add an info button
        fig = InfographicsParser._add_info_button(
            fig, plot_info, plot_info_img, image_path, plot_info_scale
        )

        return fig

    def _get_infographics(
        self,
    ) -> Figure:
        """Get the infographic for a scenario

        Returns
        -------
        Figure
            The infographic for the scenario

        """

        # Get the impact metrics
        metrics = self._get_impact_metrics()

        # Get the infographic configuration
        pie_chart_config_path = self.config_base_path.joinpath("config_charts.toml")
        pie_people_config_path = self.config_base_path.joinpath("config_people.toml")
        roads_config_path = self.config_base_path.joinpath("config_roads.toml")

        # Get the pie chart dictionaries from the configuration for charts
        return_fig = []
        try:
            charts = InfographicsParser._get_pies_dictionary(
                pie_chart_config_path, metrics
            )
            charts_fig = InfographicsParser._get_pie_chart_figure(
                data=charts.copy(),
                legend_orientation="h",
                yanchor="top",
                y=-0.1,
                image_path=self.config_base_path.joinpath("images"),
                title=charts["Other"]["Title"]["text"],
                title_font_size=charts["Other"]["Title"]["font"],
                subtitle_font_size=charts["Other"]["Subtitle"]["font"],
                image_scale=charts["Other"]["Plot"]["image_scale"],
                numbers_font=charts["Other"]["Plot"]["numbers_font"],
                legend_font_size=charts["Other"]["Legend"]["font"],
                plot_width=charts["Other"]["Plot"]["width"],
                plot_height=charts["Other"]["Plot"]["height"],
                plot_info=charts["Other"]["Info"]["text"],
                plot_info_img=charts["Other"]["Info"]["image"],
                plot_info_scale=charts["Other"]["Info"]["scale"],
            )
            return_fig.append(charts_fig)
        except FileNotFoundError:
            self.logger.warning("No charts configuration file found")

        # Get the pie chart dictionaries from the configuration for people
        try:
            people = InfographicsParser._get_pies_dictionary(
                pie_people_config_path, metrics
            )
            people_fig = InfographicsParser._get_pie_chart_figure(
                data=people.copy(),
                legend_orientation="h",
                yanchor="top",
                y=-0.1,
                image_path=self.config_base_path.joinpath("images"),
                title=people["Other"]["Title"]["text"],
                title_font_size=people["Other"]["Title"]["font"],
                subtitle_font_size=people["Other"]["Subtitle"]["font"],
                image_scale=people["Other"]["Plot"]["image_scale"],
                numbers_font=people["Other"]["Plot"]["numbers_font"],
                legend_font_size=people["Other"]["Legend"]["font"],
                plot_width=people["Other"]["Plot"]["width"],
                plot_height=people["Other"]["Plot"]["height"],
                plot_info=people["Other"]["Info"]["text"],
                plot_info_img=people["Other"]["Info"]["image"],
                plot_info_scale=people["Other"]["Info"]["scale"],
            )
            return_fig.append(people_fig)
        except FileNotFoundError:
            self.logger.warning("No people configuration file found")

        # Get the bar chart dictionaries from the configuration for roads
        try:
            roads = InfographicsParser._get_pies_dictionary(roads_config_path, metrics)
            roads_fig = InfographicsParser._get_bar_chart_figure(
                data=roads.copy(),
                image_path=self.config_base_path.joinpath("images"),
                title=roads["Other"]["Title"]["text"],
                yaxis_title = roads["Other"]["Y_axis_title"]["text"],
                title_font_size=roads["Other"]["Title"]["font"],
                subtitle_font_size=roads["Other"]["Subtitle"]["font"],
                image_scale=roads["Other"]["Plot"]["image_scale"],
                numbers_font=roads["Other"]["Plot"]["numbers_font"],
                plot_width=roads["Other"]["Plot"]["width"],
                plot_height=roads["Other"]["Plot"]["height"],
                plot_info=roads["Other"]["Info"]["text"],
                plot_info_img=roads["Other"]["Info"]["image"],
                plot_info_scale=roads["Other"]["Info"]["scale"],
            )
            return_fig.append(roads_fig)
        except FileNotFoundError:
            self.logger.warning("No roads configuration file found")

        # Return the figure
        return return_fig

    def get_infographics(self) -> Union[List[Figure], Figure]:
        """Get the infographic for a scenario

        Returns
        -------
        Union[List[Figure], Figure]
            The infographic for the scenario as a list of figures or a single figure
        """

        # Get the infographic
        infographic = self._get_infographics()

        # Return the infographic
        return infographic

    def write_infographics_to_file(self) -> str:
        """Write the infographic for a scenario to file

        Returns
        -------
        str
            The path to the infographic file
        """

        # Create the infographic path
        infographic_html = self.output_base_path.joinpath(
            f"{self.scenario_name}_metrics.html"
        )

        # Check if the infographic already exists. If so, return the path
        if Path.exists(infographic_html):
            # TODO: Print logging message
            return str(infographic_html)

        # Get the infographic
        infographic = self._get_infographics()

        # Convert the infographic to html
        self._figures_list_to_html(infographic, infographic_html)

        # Return the path to the infographic
        return str(infographic_html)

    def get_infographics_html(self) -> str:
        """Get the path to the infographic html file

        Returns
        -------
        str
            The path to the infographic html file
        """

        # Create the infographic path
        infographic_path = self.output_base_path.joinpath(
            f"{self.scenario_name}_metrics.html"
        )

        return str(infographic_path)



__version__ = "0.1.15"

from packaging import version
from pydantic import BaseModel

class FiatColumns(BaseModel):
    """
    Model defining the FIAT column types and their naming format.
    All attributes are strings that can be:
    - static: with a standard name, e.g. 'object_id'
    - dynamic: with wildcard parts, e.g. 'max_damage_{name}' or 'damage_{name}_{years}y'
    """
    object_id: str
    object_name: str
    primary_object_type: str
    secondary_object_type: str
    extraction_method: str
    ground_floor_height: str
    ground_elevation: str
    damage_function: str 
    max_potential_damage: str 
    aggregation_label: str 
    inundation_depth: str 
    inundation_depth_rp: str 
    reduction_factor: str
    reduction_factor_rp: str
    damage: str 
    damage_rp: str 
    total_damage: str 
    total_damage_rp: str 
    risk_ead: str
    segment_length: str # TODO should this be here since it is not a FIAT attribute?

def get_fiat_columns(fiat_version:str="0.2") -> FiatColumns:
    """
    Returns the column mappings for different versions of  FIAT.
    Parameters:
    fiat_version (str): The version of the FIAT. Default is "0.2".
    Returns:
    FiatColumns: An instance of FiatColumns with the appropriate column mappings for the specified version.
    Raises:
    ValueError: If the specified version is not supported.
    Supported Versions:
    - "0.2" and greater: Uses a specific set of column names.
    - "0.1.0rc2": Uses a different set of column names.
    """
    fiat_version = version.parse(fiat_version)
    # Columns for versions > 0.1
    if fiat_version > version.parse("0.1"):
        fiat_columns=FiatColumns(
        object_id = "object_id",
        object_name = "object_name",
        primary_object_type = "primary_object_type",
        secondary_object_type = "secondary_object_type",
        extraction_method = "extract_method",
        ground_floor_height = "ground_flht",
        ground_elevation = "ground_elevtn",
        damage_function = "fn_damage_{name}",
        max_potential_damage = "max_damage_{name}",
        aggregation_label = "aggregation_label:{name}",
        inundation_depth = "inun_depth",
        inundation_depth_rp = "inun_depth_{years}y",
        reduction_factor = "red_fact",
        reduction_factor_rp = "red_fact_{years}y",
        damage = "damage_{name}",
        damage_rp = "damage_{name}_{years}y",
        total_damage = "total_damage",
        total_damage_rp = "total_damage_{years}y",
        risk_ead = "ead_damage",
        segment_length = "segment_length")
    # Columns for version 0.1.0rc2
    elif fiat_version == version.parse("0.1.0rc2"):
        fiat_columns=FiatColumns(
        object_id = "Object ID",
        object_name = "Object Name",
        primary_object_type = "Primary Object Type",
        secondary_object_type = "Secondary Object Type",
        extraction_method = "Extraction Method",
        ground_floor_height = "Ground Floor Height",
        ground_elevation = "Ground Elevation",
        damage_function = "Damage Function: {name}",
        max_potential_damage = "Max Potential Damage: {name}",
        aggregation_label = "Aggregation Label: {name}",
        inundation_depth = "Inundation Depth",
        inundation_depth_rp = "Inundation Depth ({years}Y)",
        reduction_factor = "Reduction Factor",
        reduction_factor_rp = "Reduction Factor ({years}Y)",
        damage = "Damage: {name}",
        damage_rp = "Damage: {name} ({years}Y)",
        total_damage = "Total Damage",
        total_damage_rp = "Total Damage ({years}Y)",
        risk_ead = "Risk (EAD)",
        segment_length = "Segment Length")
    else:
        raise ValueError(f"Unsupported version: {fiat_version}")
    
    return fiat_columns


import os
from pathlib import Path
from typing import Union

import numpy as np
import pandas as pd
import parse

from fiat_toolbox.equity.fiat_functions import calc_rp_coef


class Equity:
    def __init__(
        self,
        census_table: Union[str, pd.DataFrame, Path],
        damages_table: Union[str, pd.DataFrame, Path],
        aggregation_label: str,
        percapitaincome_label: str,
        totalpopulation_label: str,
        damage_column_pattern: str = "Total Damage ({rp}Y)",
    ):
        """_summary_

        Parameters
        ----------
        census_table : Union[str, pd.DataFrame, Path]
            Census data
        damages_table : Union[str, pd.DataFrame, Path]
            Damage results
        aggregation_label : str
            column name of aggregation areas
        percapitaincome_label : str
            column name of per capita income
        totalpopulation_label : str
            column name of total population
        """
        # Merge tables
        self.df = self._merge_tables(census_table, damages_table, aggregation_label)
        self.df0 = self.df.copy()  # Keep copy of original
        self.aggregation_label = aggregation_label
        self.percapitaincome_label = percapitaincome_label
        self.totalpopulation_label = totalpopulation_label
        self.damage_column_pattern = damage_column_pattern

    @staticmethod
    def _check_datatype(
        variable: Union[str, pd.DataFrame],
    ) -> pd.DataFrame:
        """Check that inputs for equity are rather .csv files or pd.Dataframes

        Parameters
        ----------
        variable : Union[str, pd.DataFrame]
            input

        Returns
        -------
        pd.DataFrame
            input in dataframe format

        Raises
        ------
        ValueError
            Error if input is not in correct format
        """

        if isinstance(variable, pd.DataFrame):
            variable = variable
        elif os.path.exists(variable):
            variable = pd.read_csv(variable)
        elif isinstance(variable, str) and variable.endswith(".csv"):
            variable = pd.read_csv(variable)
        else:
            raise ValueError(
                "Input variable is neither a pandas DataFrame nor a path to a CSV file."
            )
        return variable

    @staticmethod
    def _merge_tables(
        census_table: Union[str, pd.DataFrame, Path],
        damages_table: Union[str, pd.DataFrame, Path],
        aggregation_label: str,
    ) -> pd.DataFrame:
        """Create dataframe with damage and social data used to calculate the equity weights

        Parameters
        ----------
        census_table : Union[str, pd.DataFrame, Path]
            Census data
        damages_table : Union[str, pd.DataFrame, Path]
            Damage results
        aggregation_label : str
            column name used to merge on

        Returns
        -------
        pd.DataFrame
            merged dataframe
        """
        # Check if data inputs are wether .csv files or pd.DataFrame
        census_table = Equity._check_datatype(census_table)
        damages_table = Equity._check_datatype(damages_table)
        # If the aggregated damages format is the fiat_toolbox one make sure columns are interpreted correctly
        if "Unnamed:" in damages_table.columns[0]:
            # Use name from input label
            damages_table = damages_table.rename(
                columns={damages_table.columns[0]: aggregation_label}
            )
            index_name = damages_table.columns[0]
            damages_table = damages_table.set_index(index_name)

            # Drop rows containing other variables
            damages_table = damages_table.drop(
                ["Description", "Show In Metrics Table", "Long Name"]
            )
            damages_table = damages_table.apply(pd.to_numeric)
        # Merge census block groups with fiat output (damages estimations per return period)
        df = damages_table.merge(census_table, on=aggregation_label, how="left")
        df = df[~df[aggregation_label].isna()]
        df = df.reset_index(drop=True)
        return df

    def _calculate_equity_weights(self):
        """Calculates equity weights per aggregation area"""
        # Get population and income per capital data
        I_PC = self.df[self.percapitaincome_label]  # mean per capita income
        Pop = self.df[self.totalpopulation_label]  # population

        # Calculate aggregated annual income
        I_AA = I_PC * Pop

        # Calculate weighted average income per capita
        I_PC = np.ma.MaskedArray(I_PC, mask=np.isnan(I_PC))
        I_WA = np.ma.average(I_PC, weights=Pop)

        # Calculate equity weights
        EW = (I_PC / I_WA) ** -self.gamma  # Equity Weight

        # Add annual income to the dataframe
        self.df["I_AA"] = I_AA
        # Add equity weight calculations into the dataframe
        self.df["EW"] = EW

    def _get_rp_from_name(self, name):
        parser = parse.parse(self.damage_column_pattern, name, extra_types={"s": str})
        if parser:
            rp = int(parser.named["rp"])
        else:
            rp = None
        return rp

    def calculate_ewced_per_rp(self):
        """Get equity weighted certainty equivalent damages per return period using a risk prenium"""

        # Get equity weight data
        I_AA = self.df["I_AA"]
        EW = self.df["EW"]

        # Retrieve columns with damage per return period data of fiat output
        RPs = {}
        for name in self.df.columns:
            if self._get_rp_from_name(name):
                rp = self._get_rp_from_name(name)
                RPs[rp] = name
        # Make sure data is sorted
        self.RPs = {}
        for key in sorted(RPs.keys()):
            self.RPs[key] = RPs[key]

        if len(self.RPs) == 0:
            raise ValueError(
                "Columns with damages per return period could not be found."
            )

        # Get Equity weighted certainty equivalent damage per return period
        for rp in self.RPs:
            col = self.RPs[rp]
            # Damage for return period
            D = self.df[col]
            # Period of interest in years
            t = 1
            # Probability of exceedance
            P = 1 - np.exp(-t / rp)
            # Social vulnerability
            z = D / I_AA
            # Risk premium
            R = (
                1
                - (1 + P * ((1 - z) ** (1 - self.gamma) - 1)) ** (1 / (1 - self.gamma))
            ) / (P * z)
            # This step is needed to avoid nan value when z is zero
            R[R.isna()] = 0
            # Equity weighted damage
            EWD = EW * D
            # Equity weighted certainty equivalent damage
            EWCED = R * EWD
            # Add risk premium data to dataframes
            self.df[f"R_RP_{rp}"] = R
            # Add ewd and ewced to dataframes
            self.df[f"EWD_RP_{rp}"] = EWD
            self.df[f"EWCED_RP_{rp}"] = EWCED

    def calculate_ewcead(self):
        """Calculates equity weighted certainty expected annual damages using log linear approach"""
        layers = []
        return_periods = []
        for rp in self.RPs:
            return_periods.append(rp)
            layers.append(self.df.loc[:, f"EWCED_RP_{rp}"].values)

        stacked_layers = np.dstack(tuple(layers)).squeeze()
        self.df["EWCEAD"] = stacked_layers.dot(
            np.array(calc_rp_coef(return_periods))[:, None]
        )
    
    def calculate_ewead(self):
        """Calculates equity weighted certainty expected annual damages using log linear approach"""
        layers = []
        return_periods = []
        for rp in self.RPs:
            return_periods.append(rp)
            layers.append(self.df.loc[:, f"EWD_RP_{rp}"].values)

        stacked_layers = np.dstack(tuple(layers)).squeeze()
        self.df["EWEAD"] = stacked_layers.dot(
            np.array(calc_rp_coef(return_periods))[:, None]
        )

    def equity_calculation(
        self,
        gamma: float = 1.2,
        output_file: Union[str, Path, None] = None,
    ) -> pd.DataFrame:
        """Calculates equity weighted risk

        Parameters
        ----------
        gamma : float, optional
            elasticity by default 1.2
        output_file : Union[str, Path, None], optional
            output file path, by default None

        Returns
        -------
        pd.DataFrame
            dataframe with the results
        """
        self.gamma = gamma
        # Get equity weights
        self._calculate_equity_weights()
        # Calculate equity weighted damage per return period
        self.calculate_ewced_per_rp()
        # Calculate equity weighted risk
        self.calculate_ewead()
        self.calculate_ewcead()
        # Keep only results
        df_ewced_filtered = self.df[[self.aggregation_label, "EW", "EWEAD", "EWCEAD"]]
        # Save file if requested
        if output_file is not None:
            df_ewced_filtered.to_csv(output_file, index=False)

        return df_ewced_filtered

    def rank_ewced(self, ead_column: str = "Risk (EAD)") -> pd.DataFrame:
        """Ranks areas per EAD EWCEAD and the calculates difference in ranking between 2nd and 1st

        Parameters
        ----------
        ead_column : str, optional
            name of column where the standard EAD calculation exists, by default "Risk (EAD)"

        Returns
        -------
        pd.DataFrame
            ranking results
        """
        if ead_column not in self.df.columns:
            raise ValueError(f"EAD column '{ead_column}' not present in provided aggregated file. A different column name can be specified using the 'ead_column' argument.")
        self.df["rank_EAD"] = self.df[ead_column].rank(ascending=False).astype(int)
        self.df["rank_EWEAD"] = self.df["EWEAD"].rank(ascending=False).astype(int)
        self.df["rank_EWCEAD"] = self.df["EWCEAD"].rank(ascending=False).astype(int)
        self.df["rank_diff_EWEAD"] = self.df["rank_EWEAD"] - self.df["rank_EAD"]     
        self.df["rank_diff_EWCEAD"] = self.df["rank_EWCEAD"] - self.df["rank_EAD"]
        return self.df[[self.aggregation_label, "rank_EAD", "rank_EWCEAD", "rank_diff_EWCEAD",  "rank_EWEAD", "rank_diff_EWEAD"]]

    def calculate_resilience_index(
        self, ead_column: str = "Risk (EAD)"
    ) -> pd.DataFrame:
        """Calculates a simple socioeconomic resilience indicators by the ratio of the standard EAD to the EWCEAD

        Parameters
        ----------
        ead_column : str, optional
            name of column where the standard EAD calculation exists, by default "Risk (EAD)"

        Returns
        -------
        pd.DataFrame
            index results
        """
        self.df["SRI"] = self.df[ead_column] / self.df["EWCEAD"]
        self.df = self.df.replace([np.inf, -np.inf], np.nan)
        return self.df[[self.aggregation_label, "SRI"]]



import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Union


class IInfographicsParser(ABC):
    """Interface for creating the infographic"""
    logger: logging.Logger
    
    @abstractmethod
    def __init__(
        self,
        scenario_name: str,
        metrics_full_path: Union[Path, str],
        config_base_path: Union[Path, str],
        output_base_path: Union[Path, str],
        logger: logging.Logger = logging.getLogger(__name__),
    ) -> None: ...

    @abstractmethod
    def get_infographics(self) -> str:
        """Get the infographic for a scenario

        Returns
        -------
        str
            The infographic for the scenario as a string in html format
        """
        pass

    @abstractmethod
    def write_infographics_to_file() -> str:
        """Write the infographic for a scenario to file

        Returns
        -------
        str
            The path to the infographic file
        """
        pass

    @abstractmethod
    def get_infographics_html() -> str:
        """Get the path to the infographic html file

        Returns
        -------
        str
            The path to the infographic html file
        """
        pass



import os
import re
from pathlib import Path
import shutil
import toml
import pandas as pd
import geopandas as gpd
from fiat_toolbox import get_fiat_columns

def _compile_pattern(pattern):
    """
    Compile a pattern with placeholders into a regex pattern.
    Args:
        pattern (str): The pattern containing placeholders in the format '{var}'.
    Returns:
        tuple: A tuple containing the compiled regex pattern and a list of placeholders.
    """
    # Escape special characters in pattern except for '{var}'
    escaped_pattern = re.escape(pattern)
    # Find all placeholders in the pattern
    placeholders = re.findall(r'\\{(.*?)\\}', escaped_pattern)
    # Replace placeholders with regex groups
    for placeholder in placeholders:
        escaped_pattern = escaped_pattern.replace(f'\\{{{placeholder}\\}}', f'(?P<{placeholder}>.*?)')
    # Compile the regex pattern
    regex = re.compile(f"^{escaped_pattern}$")
    return regex, placeholders

def matches_pattern(string, pattern):
    """
    Check if a string matches a pattern with placeholders.
    Args:
        string (str): The input string to be checked.
        pattern (str): The pattern containing placeholders in the format '{var}'.
    Returns:
        bool: True if the string matches the pattern, False otherwise.
    """
    regex, _ = _compile_pattern(pattern)
    return bool(regex.match(string))

def replace_pattern(string, pattern, replacement):
    """
    Replace placeholders in a string based on a pattern with a replacement string.
    Args:
        string (str): The input string to be processed.
        pattern (str): The pattern containing placeholders in the format '{var}'.
        replacement (str): The replacement string where placeholders will be replaced with corresponding values from the input string.
    Returns:
        str: The processed string with placeholders replaced by corresponding values from the input string. 
             If the pattern does not match the input string, the original string is returned.
    """
    regex, placeholders = _compile_pattern(pattern)
    
    # Find the match
    match = regex.match(string)
    if match:
        # Replace placeholders in the replacement with the captured groups
        for placeholder in placeholders:
            replacement = replacement.replace(f'{{{placeholder}}}', match.group(placeholder))
        return replacement
    return string

def convert_fiat(path_in:os.PathLike, path_out:os.PathLike, version_in:str="0.1.0rc2", version_out:str="0.2.1"):
    """
    Converts FIAT data from one version to another by copying the input directory to the output directory,
    updating the settings file, and renaming columns in the exposure CSV file according to the specified versions.
    Args:
        path_in (os.PathLike): The input directory containing the FIAT data to be converted.
        path_out (os.PathLike): The output directory where the converted FIAT data will be saved.
        version_in (str, optional): The version of the input FIAT data. Defaults to "0.1.0rc2".
        version_out (str, optional): The version of the output FIAT data. Defaults to "0.2.1".
    Raises:
        FileNotFoundError: If the settings file or exposure CSV file is not found in the input directory.
        KeyError: If the expected keys are not found in the settings file.
    """
    path_in, path_out = Path(path_in), Path(path_out)
    if path_out.exists():
        shutil.rmtree(path_out)
    shutil.copytree(path_in, path_out)
    
    settings_path = path_out.joinpath("settings.toml")
    
    with open(settings_path, 'r') as file:
        settings = toml.load(file)
    
    exposure_csv_path = settings_path.parent.joinpath(settings["exposure"]["csv"]["file"])
    exposure_csv = pd.read_csv(exposure_csv_path)
    
    format_in = get_fiat_columns(fiat_version=version_in)
    format_out = get_fiat_columns(fiat_version=version_out)
    
    name_translation = {}
    for col in exposure_csv.columns:  # iterate through output columns
        for field in list(format_out.model_fields):  # check for each field
            fiat_col = getattr(format_in , field)
            if matches_pattern(col, fiat_col):
                impact_col = getattr(format_out, field)
                new_col = replace_pattern(col, fiat_col, impact_col)
                name_translation[col] = new_col  # save mapping

    # Rename exposure csv
    exposure_csv = exposure_csv.rename(columns=name_translation)
    exposure_csv.to_csv(exposure_csv_path, index=False)
    
    
    # Get geoms
    keys = [key for key in settings["exposure"]["geom"] if "file" in key]
    geoms_paths = [settings_path.parent.joinpath(settings["exposure"]["geom"][key]) for key in keys]
    
    # Rename geoms
    for geom_path in geoms_paths:
        geom = gpd.read_file(geom_path)
        geom = geom.rename(columns=name_translation)
        geom.to_file(geom_path)


import json
import logging
import os
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Tuple, Union, Optional

import duckdb
import pandas as pd
import tomli

from fiat_toolbox.metrics_writer.fiat_metrics_interface import IMetricsFileWriter
from fiat_toolbox.metrics_writer.fiat_read_metrics_file import MetricsFileReader
from fiat_toolbox import get_fiat_columns

_AGGR_LABEL_FMT = get_fiat_columns().aggregation_label

# sql command struct
@dataclass
class sql_struct:
    name: str
    long_name: str
    show_in_metrics_table: bool
    description: str
    select: str
    filter: str
    groupby: str



class MetricsFileWriter(IMetricsFileWriter):
    """Class to parse metrics and write to a file."""
    logger: logging.Logger = logging.getLogger(__name__)
    
    def __init__(self, config_file: Union[str, Path], logger: logging.Logger = logging.getLogger(__name__), aggregation_label_fmt: str = _AGGR_LABEL_FMT):
        """
        Initialize the class.

        Parameters
        ----------
        config_file : Union[str, Path]
            The path to the metrics file.
        """
        # Convert the path to a Path object
        if isinstance(config_file, str):
            config_file = Path(config_file)

        # Check whether the file exists
        if not os.path.exists(config_file):
            raise FileNotFoundError(f"Config file '{config_file}' not found.")

        self.config_file = config_file
        self.logger = logger
        self.aggregation_label_fmt = aggregation_label_fmt
        
    def _read_metrics_file(
        self, include_aggregates: bool
    ) -> Union[Dict[str, sql_struct], Dict[str, Dict[str, sql_struct]]]:
        """
        Read a metrics file and return a list of SQL commands.

        Parameters
        ----------
        include_aggregates : bool
            Whether to include aggregation labels in the metrics.

        Returns
        -------
        Union[Dict[str, sql_struct], Dict[str, Dict[str, sql_struct]]]
            A dictionary with the SQL commands.
        """

        # Read the metrics file
        _, extension = os.path.splitext(self.config_file)
        if extension == ".json":
            metrics = json.load(open(self.config_file, "r"))
        elif extension == ".toml":
            metrics = tomli.load(open(self.config_file, "rb"))
        else:
            raise ValueError(
                f"Config file '{self.config_file}' has an invalid extension. Only .json and .toml are supported."
            )

        # Create the sql commands dictionary
        sql_command_set = {}
        if include_aggregates:
            # Check whether the metrics file contains aggregation labels
            if "aggregateBy" not in metrics or len(metrics["aggregateBy"]) == 0:
                raise ValueError(
                    "No aggregation labels specified in the metrics file, but include_aggregates is set to True."
                )
            # Loop over the aggregation labels
            for aggregate in metrics["aggregateBy"]:
                aggregate_command = {}
                # Check whether the metrics file contains metrics
                if "queries" not in metrics or len(metrics["queries"]) == 0:
                    raise ValueError("No queries specified in the metrics file.")
                # Loop over the metrics
                for metric in metrics["queries"]:
                    # Check whether the metric contains all required fields
                    if all(
                        key in metrics
                        for key in [
                            "name",
                            "long_name",
                            "show_in_metrics_table",
                            "description",
                            "select",
                            "filter",
                        ]
                    ):
                        raise ValueError(
                            f"The metrics file for metric {metric['name']} does not contain all required fields."
                        )
                    # Create the sql command
                    sql_command = sql_struct(
                        name=metric["name"],
                        long_name=metric["long_name"],
                        show_in_metrics_table=metric["show_in_metrics_table"],
                        description=metric["description"],
                        select=metric["select"],
                        filter=metric["filter"],
                        groupby=f"`{self.aggregation_label_fmt.format(name=aggregate)}`",
                    )

                    # Check whether the metric name is already in the dictionary
                    if metric["name"] in aggregate_command:
                        raise ValueError(
                            f"Duplicate metric name {metric['name']} in metrics file."
                        )

                    # Add the sql command to the dictionary
                    aggregate_command[metric["name"]] = sql_command

                # Check whether the aggregation label is already in the dictionary
                if aggregate in sql_command_set:
                    raise ValueError(
                        f"Duplicate aggregation label {aggregate} in metrics file."
                    )

                # Add the sql command to the dictionary
                sql_command_set[aggregate] = aggregate_command
        else:
            # Check whether the metrics file contains metrics
            if "queries" not in metrics or len(metrics["queries"]) == 0:
                raise ValueError("No queries specified in the metrics file.")

            # Loop over the metrics
            for metric in metrics["queries"]:
                # Check whether the metric contains all required fields
                if all(
                    key in metrics
                    for key in [
                        "name",
                        "long_name",
                        "show_in_metrics_table",
                        "description",
                        "select",
                        "filter",
                    ]
                ):
                    raise ValueError(
                        f"The metrics file for metric {metric['name']} does not contain all required fields."
                    )

                # Create the sql command
                sql_command = sql_struct(
                    name=metric["name"],
                    long_name=metric["long_name"],
                    show_in_metrics_table=metric["show_in_metrics_table"],
                    description=metric["description"],
                    select=metric["select"],
                    filter=metric["filter"],
                    groupby="",
                )

                # Check whether the metric name is already in the dictionary
                if metric["name"] in sql_command_set:
                    raise ValueError(
                        f"Duplicate metric name {metric['name']} in metrics file."
                    )

                # Add the sql command to the dictionary
                sql_command_set[metric["name"]] = sql_command

        # Return the sql commands dictionary
        return sql_command_set

    @staticmethod
    def _create_single_metric(
        df_results: pd.DataFrame, sql_command: sql_struct
    ) -> Union[Tuple[str, object], Tuple[str, Dict[str, object]]]:
        """
        Create a metrics table from the results dataframe based on an SQL command.

        Parameters
        ----------
        df_results : pd.DataFrame
            The results dataframe.
        sql_command : sql_struct
            The SQL command.

        Returns
        -------
        Union[Tuple[str, object], Tuple[str, Dict[str, object]]]
            A tuple with the metric name and value or, in the case of a groupby statement,
            a tuple with the metric name and a dictionary with the groupby variables as keys
            and the metric as value.
        """

        # First add the the groupby variables to the query
        sql_query = "SELECT "
        if sql_command.groupby:
            sql_query += f"{sql_command.groupby}, "

        # Then add the select variables
        if not sql_command.select:
            raise ValueError(
                f"No select statement specified for metric {sql_command.name}."
            )
        sql_query += f"{sql_command.select} AS `{sql_command.name}` FROM df_results"

        # Then add the filter statement
        if sql_command.filter:
            sql_query += f" WHERE {sql_command.filter}"

        # Finally add the groupby statement
        if sql_command.groupby:
            sql_query += f" GROUP BY {sql_command.groupby}"

        # Execute the query. If the query is invalid, an error PandaSQLException will be raised
        sql_query = sql_query.replace("`", '"')
        result = duckdb.query(sql_query).df()

        # If the command contains a groupby statement, return a dictionary with the groupby variables as keys and the metric as value
        if sql_command.groupby:
            # Set the groupby variables as index
            labeled_result = result.set_index(sql_command.groupby.replace("`", ""))
            # Remove rows without index name
            labeled_result = labeled_result[labeled_result.index.notna()]
            # Return the metric name and the dictionary
            return labeled_result.columns[0], dict(
                labeled_result[labeled_result.columns[0]]
            )
        # Otherwise return the metric name and the value
        return result.columns[0], result[result.columns[0]][0]

    @staticmethod
    def _create_metrics_dict(
        df_results: pd.DataFrame, sql_commands: Dict[str, sql_struct]
    ) -> Dict[str, object]:
        """
        Create a metrics table from the results dataframe based on a list of SQL commands.

        Parameters
        ----------
        df_results : pd.DataFrame
            The results dataframe.
        sql_commands : list[sql_struct]
            A list of SQL commands.

        Returns
        -------
        dict
            A dictionary with the metric names and values.
        """

        # Initialize the metrics dictionary
        df_metrics = {}

        # Run the sql commands one by one
        for name, command in sql_commands.items():
            # Create the metric (_create_single_metric is a static method, so no need to instantiate the class)
            _, value = MetricsFileWriter._create_single_metric(df_results, command)

            # Store the metric in the metrics dictionary using the metric name as key
            df_metrics[name] = value

        return df_metrics

    def _parse_metrics(
        self, df_results: pd.DataFrame, include_aggregates: bool
    ) -> Union[dict, Dict[str, dict]]:
        """
        Parse the metrics based on the config file and return a dictionary with the metrics.

        Parameters
        ----------
        df_results : pd.DataFrame
            The results dataframe.
        include_aggregates : bool
            Whether to include aggregation labels in the metrics.

        Returns
        -------
        Union[dict, List[dict]]
            A dictionary with the metrics or, in the case of multiple aggregation labels,
            a list of dictionaries.
        """

        # Read the metrics file
        sql_commands = self._read_metrics_file(include_aggregates)

        # Create the metrics dictionary
        if include_aggregates:
            metrics = {}
            # Loop over the aggregation labels
            for aggregate, commands in sql_commands.items():
                # Create the metrics dictionary for the current aggregation label (the _create_metrics_dict is a static method, so no need to instantiate the class)
                metrics[aggregate] = MetricsFileWriter._create_metrics_dict(
                    df_results, commands
                )
            return metrics
        else:
            # Create the metrics dictionary (the _create_metrics_dict is a static method, so no need to instantiate the class)
            return MetricsFileWriter._create_metrics_dict(df_results, sql_commands)

    @staticmethod
    def _write_metrics_file(
        metrics: Union[dict, Dict[str, dict]],
        config: Union[Dict[str, sql_struct], Dict[str, Dict[str, sql_struct]]],
        metrics_path: Path,
        write_aggregate: str = None,
        overwrite: bool = False,
        aggregations: list = None,
    ) -> None:
        """
        Write a metrics dictionary to a metrics file.

        Parameters
        ----------
        metrics : Union[dict, List[dict]]
            A dictionary with the metrics or, in the case of multiple aggregation labels,
            a list of dictionaries.
        config : Union[Dict[str, sql_struct], Dict[str, Dict[str, sql_struct]]]
            A dictionary with the SQL commands.
        metrics_path : Path
            The path to where to store the metrics file.
        write_aggregate : str
            The name of the aggregation label to write to the metrics file (None for no aggregation label).
        overwrite : bool
            Whether to overwrite the existing metrics file if it already exists. If False, it appends to the file.
        aggregations : list
            A list of aggregation areas. If write_aggregate is None, this is ignored.

        Returns
        -------
        None
        """

        if write_aggregate:
            # Get the metrics for the current aggregation label
            aggregate_metrics = metrics[write_aggregate]

            # Find the names dynamically
            if aggregations is None:
                aggregations = []
                for value in aggregate_metrics.values():
                    aggregations.extend(value.keys())

            # Update all empty metrics with 0
            for key, value in aggregate_metrics.items():
                if value == {}:
                    aggregate_metrics[key] = {name: 0 for name in aggregations}
                    continue
                for name in aggregations:
                    if name not in value:
                        aggregate_metrics[key][name] = 0

            # Create a dataframe from the metrics dictionary
            metricsFrame = (
                pd.DataFrame().from_dict(aggregate_metrics, orient="index").fillna(0)
            )

            # Add the long name to the dataframe
            metricsFrame.insert(
                0,
                "Long Name",
                [
                    config[write_aggregate][name].long_name
                    for name, _ in metricsFrame.iterrows()
                ],
            )

            # Add the metrics table selector to the dataframe
            metricsFrame.insert(
                0,
                "Show In Metrics Table",
                [
                    config[write_aggregate][name].show_in_metrics_table
                    for name, _ in metricsFrame.iterrows()
                ],
            )

            # Add the description to the dataframe
            metricsFrame.insert(
                0,
                "Description",
                [
                    config[write_aggregate][name].description
                    for name, _ in metricsFrame.iterrows()
                ],
            )

            # Check if the file already exists
            if os.path.exists(metrics_path):
                if overwrite:
                    MetricsFileWriter.logger.warning(
                        f"Metrics file '{metrics_path}' already exists. Overwriting..."
                    )
                    os.remove(metrics_path)
                else:
                    new_metrics = MetricsFileReader(
                        metrics_path
                    ).read_metrics_from_file(
                        include_long_names=True,
                        include_description=True,
                        include_metrics_table_selection=True,
                    )
                    metricsFrame = pd.concat([new_metrics, metricsFrame])

            # Transpose the dataframe
            metricsFrame = metricsFrame.transpose()

            # Write the metrics to a file
            if metrics_path.parent and not metrics_path.parent.exists():
                metrics_path.parent.mkdir(parents=True)
            metricsFrame.to_csv(metrics_path)
        else:
            # Create a dataframe from the metrics dictionary
            metricsFrame = (
                pd.DataFrame()
                .from_dict(metrics, orient="index", columns=["Value"])
                .fillna(0)
            )

            # Add the long name to the dataframe
            metricsFrame.insert(
                0,
                "Long Name",
                [config[name].long_name for name, _ in metricsFrame.iterrows()],
            )

            # Add the metrics table selector to the dataframe
            metricsFrame.insert(
                0,
                "Show In Metrics Table",
                [
                    config[name].show_in_metrics_table
                    for name, _ in metricsFrame.iterrows()
                ],
            )

            # Add the description to the dataframe
            metricsFrame.insert(
                0,
                "Description",
                [config[name].description for name, _ in metricsFrame.iterrows()],
            )

            # Check if the file already exists
            if os.path.exists(metrics_path):
                if overwrite:
                    logging.warning(
                        f"Metrics file '{metrics_path}' already exists. Overwriting..."
                    )
                    os.remove(metrics_path)
                else:
                    new_metrics = MetricsFileReader(
                        metrics_path
                    ).read_metrics_from_file(
                        include_long_names=True,
                        include_description=True,
                        include_metrics_table_selection=True,
                    )
                    metricsFrame = pd.concat([new_metrics, metricsFrame])

            # Write the metrics to a file
            if metrics_path.parent and not metrics_path.parent.exists():
                metrics_path.parent.mkdir(parents=True)
            metricsFrame.to_csv(metrics_path)

    def parse_metrics_to_file(
        self,
        df_results: pd.DataFrame,
        metrics_path: Union[str, Path],
        write_aggregate: str = None,
        overwrite: bool = False,
    ) -> Union[str, Dict[str, str]]:
        """
        Parse a metrics file and write the metrics to a file.

        Parameters
        ----------
        df_results : pd.DataFrame
            The results dataframe.
        metrics_path : Union[str, Path]
            The path to where to store the metrics file.
        write_aggregate : str
            The name of the aggregation label to write to the metrics file
            (None for no aggregation label, 'all' for all possible ones).
        overwrite : bool
            Whether to overwrite the existing metrics file if it already exists. If False, it appends to the file.

        Returns
        -------
        Union[str, Dict[str, str]]
            The path to the metrics file or a dictionary with the aggregation labels as keys
            and the paths to the metrics files as values.
        """

        # Convert the path to a Path object
        if isinstance(metrics_path, str):
            metrics_path = Path(metrics_path)

        # Check whether to include aggregation labels
        include_aggregates = True if write_aggregate else False

        # Read the metrics config file
        config = self._read_metrics_file(include_aggregates)

        # Parse the metrics
        metrics = self._parse_metrics(df_results, include_aggregates)

        # Write the metrics to a file
        if write_aggregate == "all":
            # Initialize the return dictionary
            return_files = {}
            for key in config.keys():
                # If using aggregation labels, add the aggregation label to the filename
                directory, filename = os.path.split(metrics_path)
                filename, extension = os.path.splitext(filename)
                new_filename = filename + "_" + key + extension
                new_path = Path(os.path.join(directory, new_filename))
                return_files[key] = new_path

                # Write the metrics to a file
                MetricsFileWriter._write_metrics_file(
                    metrics,
                    config,
                    new_path,
                    write_aggregate=key,
                    overwrite=overwrite,
                    aggregations=df_results[self.aggregation_label_fmt.format(name=key)].unique(),
                )
        else:
            # Write the metrics to a file
            MetricsFileWriter._write_metrics_file(
                metrics,
                config,
                metrics_path,
                write_aggregate=write_aggregate,
                overwrite=overwrite,
            )
            return_files = metrics_path

        return return_files
